/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(GPU_PROCESS)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "GraphicsContextGLIdentifier.h"
#include "MessageNames.h"
#include "RemoteAudioHardwareListenerIdentifier.h"
#include "RemoteRemoteCommandListenerIdentifier.h"
#include "RenderingBackendIdentifier.h"
#include "SampleBufferDisplayLayerIdentifier.h"
#include "StreamServerConnection.h"
#if USE(GRAPHICS_LAYER_WC)
#include "WCLayerTreeHostIdentifier.h"
#endif
#include "WebGPUIdentifier.h"
#include "WebPageProxyIdentifier.h"
#include <WebCore/GraphicsContextGLAttributes.h>
#include <WebCore/PageIdentifier.h>
#include <optional>
#include <wtf/Forward.h>
#include <wtf/MachSendRight.h>
#include <wtf/RuntimeApplicationChecks.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class FloatRect;
struct NowPlayingInfo;
}

namespace WebKit {
struct MediaOverridesForTesting;
struct RemoteAudioSessionConfiguration;
}

namespace Messages {
namespace GPUConnectionToWebProcess {

static inline IPC::ReceiverName messageReceiverName()
{
#if ASSERT_ENABLED
    static std::once_flag onceFlag;
    std::call_once(
        onceFlag,
        [&] {
            ASSERT(isInGPUProcess());
        }
    );
#endif
    return IPC::ReceiverName::GPUConnectionToWebProcess;
}

class CreateRenderingBackend {
public:
    using Arguments = std::tuple<WebKit::RenderingBackendIdentifier, IPC::StreamServerConnectionHandle>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateRenderingBackend; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CreateRenderingBackend(const WebKit::RenderingBackendIdentifier& renderingBackendIdentifier, IPC::StreamServerConnectionHandle&& connectionHandle)
        : m_arguments(renderingBackendIdentifier, WTFMove(connectionHandle))
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RenderingBackendIdentifier&, IPC::StreamServerConnectionHandle&&> m_arguments;
};

class ReleaseRenderingBackend {
public:
    using Arguments = std::tuple<WebKit::RenderingBackendIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseRenderingBackend; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseRenderingBackend(const WebKit::RenderingBackendIdentifier& renderingBackendIdentifier)
        : m_arguments(renderingBackendIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RenderingBackendIdentifier&> m_arguments;
};

#if ENABLE(WEBGL)
class CreateGraphicsContextGL {
public:
    using Arguments = std::tuple<WebKit::GraphicsContextGLIdentifier, WebCore::GraphicsContextGLAttributes, WebKit::RenderingBackendIdentifier, IPC::StreamServerConnectionHandle>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateGraphicsContextGL; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CreateGraphicsContextGL(const WebKit::GraphicsContextGLIdentifier& identifier, const WebCore::GraphicsContextGLAttributes& attributes, const WebKit::RenderingBackendIdentifier& renderingBackendIdentifier, IPC::StreamServerConnectionHandle&& serverConnection)
        : m_arguments(identifier, attributes, renderingBackendIdentifier, WTFMove(serverConnection))
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::GraphicsContextGLIdentifier&, const WebCore::GraphicsContextGLAttributes&, const WebKit::RenderingBackendIdentifier&, IPC::StreamServerConnectionHandle&&> m_arguments;
};
#endif

#if ENABLE(WEBGL)
class ReleaseGraphicsContextGL {
public:
    using Arguments = std::tuple<WebKit::GraphicsContextGLIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseGraphicsContextGL; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseGraphicsContextGL(const WebKit::GraphicsContextGLIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::GraphicsContextGLIdentifier&> m_arguments;
};
#endif

class CreateGPU {
public:
    using Arguments = std::tuple<WebKit::WebGPUIdentifier, WebKit::RenderingBackendIdentifier, IPC::StreamServerConnectionHandle>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateGPU; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CreateGPU(const WebKit::WebGPUIdentifier& identifier, const WebKit::RenderingBackendIdentifier& renderingBackendIdentifier, IPC::StreamServerConnectionHandle&& serverConnection)
        : m_arguments(identifier, renderingBackendIdentifier, WTFMove(serverConnection))
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WebGPUIdentifier&, const WebKit::RenderingBackendIdentifier&, IPC::StreamServerConnectionHandle&&> m_arguments;
};

class ReleaseGPU {
public:
    using Arguments = std::tuple<WebKit::WebGPUIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseGPU; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseGPU(const WebKit::WebGPUIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WebGPUIdentifier&> m_arguments;
};

class ClearNowPlayingInfo {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ClearNowPlayingInfo; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ClearNowPlayingInfo()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class SetNowPlayingInfo {
public:
    using Arguments = std::tuple<WebCore::NowPlayingInfo>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_SetNowPlayingInfo; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetNowPlayingInfo(const WebCore::NowPlayingInfo& nowPlayingInfo)
        : m_arguments(nowPlayingInfo)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::NowPlayingInfo&> m_arguments;
};

#if USE(AUDIO_SESSION)
class EnsureAudioSession {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_EnsureAudioSession; }
    static constexpr bool isSync = true;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<WebKit::RemoteAudioSessionConfiguration>;
    using Reply = CompletionHandler<void(WebKit::RemoteAudioSessionConfiguration&&)>;
    EnsureAudioSession()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class EnsureMediaSessionHelper {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_EnsureMediaSessionHelper; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    EnsureMediaSessionHelper()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if HAVE(VISIBILITY_PROPAGATION_VIEW)
class CreateVisibilityPropagationContextForPage {
public:
    using Arguments = std::tuple<WebKit::WebPageProxyIdentifier, WebCore::PageIdentifier, bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateVisibilityPropagationContextForPage; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CreateVisibilityPropagationContextForPage(const WebKit::WebPageProxyIdentifier& pageProxyID, const WebCore::PageIdentifier& pageID, bool canShowWhileLocked)
        : m_arguments(pageProxyID, pageID, canShowWhileLocked)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WebPageProxyIdentifier&, const WebCore::PageIdentifier&, bool> m_arguments;
};
#endif

#if HAVE(VISIBILITY_PROPAGATION_VIEW)
class DestroyVisibilityPropagationContextForPage {
public:
    using Arguments = std::tuple<WebKit::WebPageProxyIdentifier, WebCore::PageIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_DestroyVisibilityPropagationContextForPage; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    DestroyVisibilityPropagationContextForPage(const WebKit::WebPageProxyIdentifier& pageProxyID, const WebCore::PageIdentifier& pageID)
        : m_arguments(pageProxyID, pageID)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WebPageProxyIdentifier&, const WebCore::PageIdentifier&> m_arguments;
};
#endif

class SetMediaOverridesForTesting {
public:
    using Arguments = std::tuple<WebKit::MediaOverridesForTesting>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_SetMediaOverridesForTesting; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetMediaOverridesForTesting(const WebKit::MediaOverridesForTesting& configuration)
        : m_arguments(configuration)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::MediaOverridesForTesting&> m_arguments;
};

class CreateAudioHardwareListener {
public:
    using Arguments = std::tuple<WebKit::RemoteAudioHardwareListenerIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateAudioHardwareListener; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit CreateAudioHardwareListener(const WebKit::RemoteAudioHardwareListenerIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteAudioHardwareListenerIdentifier&> m_arguments;
};

class ReleaseAudioHardwareListener {
public:
    using Arguments = std::tuple<WebKit::RemoteAudioHardwareListenerIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseAudioHardwareListener; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseAudioHardwareListener(const WebKit::RemoteAudioHardwareListenerIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteAudioHardwareListenerIdentifier&> m_arguments;
};

class CreateRemoteCommandListener {
public:
    using Arguments = std::tuple<WebKit::RemoteRemoteCommandListenerIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateRemoteCommandListener; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit CreateRemoteCommandListener(const WebKit::RemoteRemoteCommandListenerIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteRemoteCommandListenerIdentifier&> m_arguments;
};

class ReleaseRemoteCommandListener {
public:
    using Arguments = std::tuple<WebKit::RemoteRemoteCommandListenerIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseRemoteCommandListener; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseRemoteCommandListener(const WebKit::RemoteRemoteCommandListenerIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteRemoteCommandListenerIdentifier&> m_arguments;
};

class ConfigureLoggingChannel {
public:
    using Arguments = std::tuple<String, WTFLogChannelState, WTFLogLevel>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ConfigureLoggingChannel; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ConfigureLoggingChannel(const String& channelName, WTFLogChannelState state, WTFLogLevel level)
        : m_arguments(channelName, state, level)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const String&, WTFLogChannelState, WTFLogLevel> m_arguments;
};

#if USE(GRAPHICS_LAYER_WC)
class CreateWCLayerTreeHost {
public:
    using Arguments = std::tuple<WebKit::WCLayerTreeHostIdentifier, uint64_t, bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_CreateWCLayerTreeHost; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CreateWCLayerTreeHost(const WebKit::WCLayerTreeHostIdentifier& identifier, uint64_t nativeWindow, bool usesOffscreenRendering)
        : m_arguments(identifier, nativeWindow, usesOffscreenRendering)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WCLayerTreeHostIdentifier&, uint64_t, bool> m_arguments;
};
#endif

#if USE(GRAPHICS_LAYER_WC)
class ReleaseWCLayerTreeHost {
public:
    using Arguments = std::tuple<WebKit::WCLayerTreeHostIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_ReleaseWCLayerTreeHost; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ReleaseWCLayerTreeHost(const WebKit::WCLayerTreeHostIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::WCLayerTreeHostIdentifier&> m_arguments;
};
#endif

#if ENABLE(MEDIA_SOURCE)
class EnableMockMediaSource {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_EnableMockMediaSource; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    EnableMockMediaSource()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if PLATFORM(COCOA) && ENABLE(MEDIA_STREAM)
class UpdateSampleBufferDisplayLayerBoundsAndPosition {
public:
    using Arguments = std::tuple<WebKit::SampleBufferDisplayLayerIdentifier, WebCore::FloatRect, std::optional<MachSendRight>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_UpdateSampleBufferDisplayLayerBoundsAndPosition; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    UpdateSampleBufferDisplayLayerBoundsAndPosition(const WebKit::SampleBufferDisplayLayerIdentifier& identifier, const WebCore::FloatRect& bounds, std::optional<MachSendRight>&& fence)
        : m_arguments(identifier, bounds, WTFMove(fence))
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::SampleBufferDisplayLayerIdentifier&, const WebCore::FloatRect&, std::optional<MachSendRight>&&> m_arguments;
};
#endif

#if ENABLE(EXTENSION_CAPABILITIES)
class SetMediaEnvironment {
public:
    using Arguments = std::tuple<WebCore::PageIdentifier, String>;

    static IPC::MessageName name() { return IPC::MessageName::GPUConnectionToWebProcess_SetMediaEnvironment; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    SetMediaEnvironment(const WebCore::PageIdentifier& pageIdentifier, const String& mediaEnvironment)
        : m_arguments(pageIdentifier, mediaEnvironment)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::PageIdentifier&, const String&> m_arguments;
};
#endif

} // namespace GPUConnectionToWebProcess
} // namespace Messages

#endif // ENABLE(GPU_PROCESS)
