/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "GPUProcessConnection.h"

#include "ArgumentCoders.h" // NOLINT
#if ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
#include "AudioMediaStreamTrackRendererInternalUnitIdentifier.h" // NOLINT
#endif
#include "Decoder.h" // NOLINT
#include "GPUProcessConnectionInfo.h" // NOLINT
#include "GPUProcessConnectionMessages.h" // NOLINT
#include "HandleMessage.h" // NOLINT
#if ENABLE(ROUTING_ARBITRATION)
#include <WebCore/AudioSession.h> // NOLINT
#endif
#include <WebCore/PlatformMediaSession.h> // NOLINT
#include <optional> // NOLINT

#if ENABLE(IPC_TESTING_API)
#include "JSIPCBinding.h"
#endif

namespace WebKit {

void GPUProcessConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::GPUProcessConnection::DidInitialize::name())
        return IPC::handleMessage<Messages::GPUProcessConnection::DidInitialize>(connection, decoder, this, &GPUProcessConnection::didInitialize);
    if (decoder.messageName() == Messages::GPUProcessConnection::DidReceiveRemoteCommand::name())
        return IPC::handleMessage<Messages::GPUProcessConnection::DidReceiveRemoteCommand>(connection, decoder, this, &GPUProcessConnection::didReceiveRemoteCommand);
#if ENABLE(ROUTING_ARBITRATION)
    if (decoder.messageName() == Messages::GPUProcessConnection::BeginRoutingArbitrationWithCategory::name())
        return IPC::handleMessageAsync<Messages::GPUProcessConnection::BeginRoutingArbitrationWithCategory>(connection, decoder, this, &GPUProcessConnection::beginRoutingArbitrationWithCategory);
    if (decoder.messageName() == Messages::GPUProcessConnection::EndRoutingArbitration::name())
        return IPC::handleMessage<Messages::GPUProcessConnection::EndRoutingArbitration>(connection, decoder, this, &GPUProcessConnection::endRoutingArbitration);
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    if (decoder.messageName() == Messages::GPUProcessConnection::ResetAudioMediaStreamTrackRendererInternalUnit::name())
        return IPC::handleMessage<Messages::GPUProcessConnection::ResetAudioMediaStreamTrackRendererInternalUnit>(connection, decoder, this, &GPUProcessConnection::resetAudioMediaStreamTrackRendererInternalUnit);
#endif
    if (dispatchMessage(connection, decoder))
        return;
    UNUSED_PARAM(connection);
    RELEASE_LOG_ERROR(IPC, "Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()).characters(), decoder.destinationID());
    decoder.markInvalid();
}

bool GPUProcessConnection::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, UniqueRef<IPC::Encoder>& replyEncoder)
{
    ASSERT(isInWebProcess());
    Ref protectedThis { *this };
    if (dispatchSyncMessage(connection, decoder, replyEncoder))
        return true;
    UNUSED_PARAM(connection);
    UNUSED_PARAM(replyEncoder);
    RELEASE_LOG_ERROR(IPC, "Unhandled synchronous message %s to %" PRIu64, description(decoder.messageName()).characters(), decoder.destinationID());
    decoder.markInvalid();
    return false;
}

} // namespace WebKit

#if ENABLE(IPC_TESTING_API)

namespace IPC {

template<> std::optional<JSC::JSValue> jsValueForDecodedMessage<MessageName::GPUProcessConnection_DidInitialize>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::DidInitialize::Arguments>(globalObject, decoder);
}
template<> std::optional<JSC::JSValue> jsValueForDecodedMessage<MessageName::GPUProcessConnection_DidReceiveRemoteCommand>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::DidReceiveRemoteCommand::Arguments>(globalObject, decoder);
}
#if ENABLE(ROUTING_ARBITRATION)
template<> std::optional<JSC::JSValue> jsValueForDecodedMessage<MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::BeginRoutingArbitrationWithCategory::Arguments>(globalObject, decoder);
}
template<> std::optional<JSC::JSValue> jsValueForDecodedMessageReply<MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::BeginRoutingArbitrationWithCategory::ReplyArguments>(globalObject, decoder);
}
template<> std::optional<JSC::JSValue> jsValueForDecodedMessage<MessageName::GPUProcessConnection_EndRoutingArbitration>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::EndRoutingArbitration::Arguments>(globalObject, decoder);
}
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
template<> std::optional<JSC::JSValue> jsValueForDecodedMessage<MessageName::GPUProcessConnection_ResetAudioMediaStreamTrackRendererInternalUnit>(JSC::JSGlobalObject* globalObject, Decoder& decoder)
{
    return jsValueForDecodedArguments<Messages::GPUProcessConnection::ResetAudioMediaStreamTrackRendererInternalUnit::Arguments>(globalObject, decoder);
}
#endif

}

#endif


#endif // ENABLE(GPU_PROCESS)
