/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(GPU_PROCESS)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "ConnectionHandle.h"
#if HAVE(AUDIT_TOKEN)
#include "CoreIPCAuditToken.h"
#endif
#include "GPUProcessConnectionParameters.h"
#include "GPUProcessCreationParameters.h"
#include "GPUProcessSessionParameters.h"
#include "MessageNames.h"
#include "SandboxExtension.h"
#include "SharedFileHandle.h"
#include <WebCore/CaptureDevice.h>
#include <WebCore/IntDegrees.h>
#include <WebCore/MediaPlayerIdentifier.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/ProcessIdentifier.h>
#include <WebCore/ProcessIdentity.h>
#include <WebCore/ShareableBitmapHandle.h>
#include <optional>
#include <span>
#include <wtf/Forward.h>
#include <wtf/RuntimeApplicationChecks.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace IPC {
class SharedBufferReference;
}

namespace PAL {
class SessionID;
}

namespace WebCore {
class SecurityOriginData;
enum class VideoFrameRotation : uint16_t;
enum class DisplayCapturePromptType : uint8_t;
struct MockMediaDevice;
struct ScreenProperties;
}

namespace WebKit {
struct CoreIPCAuditToken;
struct GPUProcessPreferences;
struct SharedPreferencesForWebProcess;
}

namespace Messages {
namespace GPUProcess {

static inline IPC::ReceiverName messageReceiverName()
{
#if ASSERT_ENABLED
    static std::once_flag onceFlag;
    std::call_once(
        onceFlag,
        [&] {
            ASSERT(isInGPUProcess());
        }
    );
#endif
    return IPC::ReceiverName::GPUProcess;
}

class InitializeGPUProcess {
public:
    using Arguments = std::tuple<WebKit::GPUProcessCreationParameters>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_InitializeGPUProcess; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_InitializeGPUProcessReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    explicit InitializeGPUProcess(WebKit::GPUProcessCreationParameters&& processCreationParameters)
        : m_arguments(WTFMove(processCreationParameters))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebKit::GPUProcessCreationParameters&&> m_arguments;
};

class CreateGPUConnectionToWebProcess {
public:
    using Arguments = std::tuple<WebCore::ProcessIdentifier, PAL::SessionID, IPC::ConnectionHandle, WebKit::GPUProcessConnectionParameters>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_CreateGPUConnectionToWebProcess; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_CreateGPUConnectionToWebProcessReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    CreateGPUConnectionToWebProcess(const WebCore::ProcessIdentifier& processIdentifier, const PAL::SessionID& sessionID, IPC::ConnectionHandle&& connectionHandle, WebKit::GPUProcessConnectionParameters&& parameters)
        : m_arguments(processIdentifier, sessionID, WTFMove(connectionHandle), WTFMove(parameters))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::ProcessIdentifier&, const PAL::SessionID&, IPC::ConnectionHandle&&, WebKit::GPUProcessConnectionParameters&&> m_arguments;
};

class SharedPreferencesForWebProcessDidChange {
public:
    using Arguments = std::tuple<WebCore::ProcessIdentifier, WebKit::SharedPreferencesForWebProcess>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SharedPreferencesForWebProcessDidChange; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_SharedPreferencesForWebProcessDidChangeReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    SharedPreferencesForWebProcessDidChange(const WebCore::ProcessIdentifier& processIdentifier, const WebKit::SharedPreferencesForWebProcess& sharedPreferencesForWebProcess)
        : m_arguments(processIdentifier, sharedPreferencesForWebProcess)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::ProcessIdentifier&, const WebKit::SharedPreferencesForWebProcess&> m_arguments;
};

class UpdateGPUProcessPreferences {
public:
    using Arguments = std::tuple<WebKit::GPUProcessPreferences>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_UpdateGPUProcessPreferences; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit UpdateGPUProcessPreferences(const WebKit::GPUProcessPreferences& preferences)
        : m_arguments(preferences)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::GPUProcessPreferences&> m_arguments;
};

class UpdateSandboxAccess {
public:
    using Arguments = std::tuple<Vector<WebKit::SandboxExtensionHandle>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_UpdateSandboxAccess; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit UpdateSandboxAccess(Vector<WebKit::SandboxExtensionHandle>&& extensions)
        : m_arguments(WTFMove(extensions))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<Vector<WebKit::SandboxExtensionHandle>&&> m_arguments;
};

class PrepareToSuspend {
public:
    using Arguments = std::tuple<bool, MonotonicTime>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_PrepareToSuspend; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_PrepareToSuspendReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    PrepareToSuspend(bool isSuspensionImminent, const MonotonicTime& estimatedSuspendTime)
        : m_arguments(isSuspensionImminent, estimatedSuspendTime)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool, const MonotonicTime&> m_arguments;
};

class ProcessDidResume {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_ProcessDidResume; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ProcessDidResume()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class AddSession {
public:
    using Arguments = std::tuple<PAL::SessionID, WebKit::GPUProcessSessionParameters>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_AddSession; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    AddSession(const PAL::SessionID& sessionID, WebKit::GPUProcessSessionParameters&& parameters)
        : m_arguments(sessionID, WTFMove(parameters))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const PAL::SessionID&, WebKit::GPUProcessSessionParameters&&> m_arguments;
};

class RemoveSession {
public:
    using Arguments = std::tuple<PAL::SessionID>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_RemoveSession; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveSession(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const PAL::SessionID&> m_arguments;
};

#if ENABLE(MEDIA_STREAM)
class SetMockCaptureDevicesEnabled {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetMockCaptureDevicesEnabled; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetMockCaptureDevicesEnabled(bool isEnabled)
        : m_arguments(isEnabled)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class SetOrientationForMediaCapture {
public:
    using Arguments = std::tuple<WebCore::IntDegrees>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetOrientationForMediaCapture; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetOrientationForMediaCapture(const WebCore::IntDegrees& orientation)
        : m_arguments(orientation)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::IntDegrees&> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class RotationAngleForCaptureDeviceChanged {
public:
    using Arguments = std::tuple<String, WebCore::VideoFrameRotation>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_RotationAngleForCaptureDeviceChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RotationAngleForCaptureDeviceChanged(const String& persistentId, WebCore::VideoFrameRotation rotation)
        : m_arguments(persistentId, rotation)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const String&, WebCore::VideoFrameRotation> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class UpdateCaptureAccess {
public:
    using Arguments = std::tuple<bool, bool, bool, WebCore::ProcessIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_UpdateCaptureAccess; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_UpdateCaptureAccessReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    UpdateCaptureAccess(bool allowAudioCapture, bool allowVideoCapture, bool allowDisplayCapture, const WebCore::ProcessIdentifier& processID)
        : m_arguments(allowAudioCapture, allowVideoCapture, allowDisplayCapture, processID)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool, bool, bool, const WebCore::ProcessIdentifier&> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class UpdateCaptureOrigin {
public:
    using Arguments = std::tuple<WebCore::SecurityOriginData, WebCore::ProcessIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_UpdateCaptureOrigin; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    UpdateCaptureOrigin(const WebCore::SecurityOriginData& originData, const WebCore::ProcessIdentifier& processID)
        : m_arguments(originData, processID)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::SecurityOriginData&, const WebCore::ProcessIdentifier&> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class AddMockMediaDevice {
public:
    using Arguments = std::tuple<WebCore::MockMediaDevice>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_AddMockMediaDevice; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit AddMockMediaDevice(const WebCore::MockMediaDevice& device)
        : m_arguments(device)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::MockMediaDevice&> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class ClearMockMediaDevices {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_ClearMockMediaDevices; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ClearMockMediaDevices()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class RemoveMockMediaDevice {
public:
    using Arguments = std::tuple<String>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_RemoveMockMediaDevice; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveMockMediaDevice(const String& persistentId)
        : m_arguments(persistentId)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const String&> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class SetMockMediaDeviceIsEphemeral {
public:
    using Arguments = std::tuple<String, bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetMockMediaDeviceIsEphemeral; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    SetMockMediaDeviceIsEphemeral(const String& persistentId, bool isEphemeral)
        : m_arguments(persistentId, isEphemeral)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const String&, bool> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class ResetMockMediaDevices {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_ResetMockMediaDevices; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ResetMockMediaDevices()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class SetMockCaptureDevicesInterrupted {
public:
    using Arguments = std::tuple<bool, bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetMockCaptureDevicesInterrupted; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    SetMockCaptureDevicesInterrupted(bool isCameraInterrupted, bool isMicrophoneInterrupted)
        : m_arguments(isCameraInterrupted, isMicrophoneInterrupted)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool, bool> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class TriggerMockCaptureConfigurationChange {
public:
    using Arguments = std::tuple<bool, bool, bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_TriggerMockCaptureConfigurationChange; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    TriggerMockCaptureConfigurationChange(bool forCamera, bool forMicrophone, bool forDisplay)
        : m_arguments(forCamera, forMicrophone, forDisplay)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool, bool, bool> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class SetShouldListenToVoiceActivity {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetShouldListenToVoiceActivity; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetShouldListenToVoiceActivity(bool shouldListen)
        : m_arguments(shouldListen)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class EnableMicrophoneMuteStatusAPI {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_EnableMicrophoneMuteStatusAPI; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    EnableMicrophoneMuteStatusAPI()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if HAVE(SC_CONTENT_SHARING_PICKER)
class SetUseSCContentSharingPicker {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetUseSCContentSharingPicker; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetUseSCContentSharingPicker(bool use)
        : m_arguments(use)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};
#endif

#if PLATFORM(MAC)
class SetScreenProperties {
public:
    using Arguments = std::tuple<WebCore::ScreenProperties>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetScreenProperties; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SetScreenProperties(const WebCore::ScreenProperties& screenProperties)
        : m_arguments(screenProperties)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::ScreenProperties&> m_arguments;
};
#endif

#if HAVE(SCREEN_CAPTURE_KIT)
class PromptForGetDisplayMedia {
public:
    using Arguments = std::tuple<WebCore::DisplayCapturePromptType>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_PromptForGetDisplayMedia; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_PromptForGetDisplayMediaReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<std::optional<WebCore::CaptureDevice>>;
    using Reply = CompletionHandler<void(std::optional<WebCore::CaptureDevice>&&)>;
    using Promise = WTF::NativePromise<std::optional<WebCore::CaptureDevice>, IPC::Error>;
    explicit PromptForGetDisplayMedia(WebCore::DisplayCapturePromptType type)
        : m_arguments(type)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::DisplayCapturePromptType> m_arguments;
};
#endif

#if HAVE(SCREEN_CAPTURE_KIT)
class CancelGetDisplayMediaPrompt {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_CancelGetDisplayMediaPrompt; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    CancelGetDisplayMediaPrompt()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};
#endif

#if ENABLE(VIDEO)
class RequestBitmapImageForCurrentTime {
public:
    using Arguments = std::tuple<WebCore::ProcessIdentifier, WebCore::MediaPlayerIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_RequestBitmapImageForCurrentTime; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_RequestBitmapImageForCurrentTimeReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<std::optional<WebCore::ShareableBitmapHandle>>;
    using Reply = CompletionHandler<void(std::optional<WebCore::ShareableBitmapHandle>&&)>;
    using Promise = WTF::NativePromise<std::optional<WebCore::ShareableBitmapHandle>, IPC::Error>;
    RequestBitmapImageForCurrentTime(const WebCore::ProcessIdentifier& processIdentifier, const WebCore::MediaPlayerIdentifier& playerIdentifier)
        : m_arguments(processIdentifier, playerIdentifier)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::ProcessIdentifier&, const WebCore::MediaPlayerIdentifier&> m_arguments;
};
#endif

#if PLATFORM(MAC)
class OpenDirectoryCacheInvalidated {
public:
    using Arguments = std::tuple<WebKit::SandboxExtensionHandle>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_OpenDirectoryCacheInvalidated; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit OpenDirectoryCacheInvalidated(WebKit::SandboxExtensionHandle&& handle)
        : m_arguments(WTFMove(handle))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebKit::SandboxExtensionHandle&&> m_arguments;
};
#endif

#if HAVE(AUDIO_COMPONENT_SERVER_REGISTRATIONS)
class ConsumeAudioComponentRegistrations {
public:
    using Arguments = std::tuple<IPC::SharedBufferReference>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_ConsumeAudioComponentRegistrations; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ConsumeAudioComponentRegistrations(const IPC::SharedBufferReference& registrationData)
        : m_arguments(registrationData)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const IPC::SharedBufferReference&> m_arguments;
};
#endif

#if HAVE(POWERLOG_TASK_MODE_QUERY)
class EnablePowerLogging {
public:
    using Arguments = std::tuple<WebKit::SandboxExtensionHandle>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_EnablePowerLogging; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit EnablePowerLogging(WebKit::SandboxExtensionHandle&& handle)
        : m_arguments(WTFMove(handle))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebKit::SandboxExtensionHandle&&> m_arguments;
};
#endif

class UserPreferredLanguagesChanged {
public:
    using Arguments = std::tuple<Vector<String>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_UserPreferredLanguagesChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit UserPreferredLanguagesChanged(const Vector<String>& languages)
        : m_arguments(languages)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Vector<String>&> m_arguments;
};

class WebProcessConnectionCountForTesting {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_WebProcessConnectionCountForTesting; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_WebProcessConnectionCountForTestingReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<uint64_t>;
    using Reply = CompletionHandler<void(uint64_t)>;
    using Promise = WTF::NativePromise<uint64_t, IPC::Error>;
    WebProcessConnectionCountForTesting()
        : m_arguments()
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

#if USE(EXTENSIONKIT)
class ResolveBookmarkDataForCacheDirectory {
public:
    using Arguments = std::tuple<std::span<const uint8_t>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_ResolveBookmarkDataForCacheDirectory; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit ResolveBookmarkDataForCacheDirectory(const std::span<const uint8_t>& bookmarkData)
        : m_arguments(bookmarkData)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const std::span<const uint8_t>&> m_arguments;
};
#endif

#if ENABLE(WEBXR)
class WebXRPromptAccepted {
public:
    using Arguments = std::tuple<std::optional<WebCore::ProcessIdentity>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_WebXRPromptAccepted; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_WebXRPromptAcceptedReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<bool>;
    using Reply = CompletionHandler<void(bool)>;
    using Promise = WTF::NativePromise<bool, IPC::Error>;
    explicit WebXRPromptAccepted(std::optional<WebCore::ProcessIdentity>&& processIdentity)
        : m_arguments(WTFMove(processIdentity))
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<std::optional<WebCore::ProcessIdentity>&&> m_arguments;
};
#endif

#if HAVE(AUDIT_TOKEN)
class SetPresentingApplicationAuditToken {
public:
    using Arguments = std::tuple<WebCore::ProcessIdentifier, WebCore::PageIdentifier, std::optional<WebKit::CoreIPCAuditToken>>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_SetPresentingApplicationAuditToken; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    SetPresentingApplicationAuditToken(const WebCore::ProcessIdentifier& processIdentifier, const WebCore::PageIdentifier& pageIdentifier, const std::optional<WebKit::CoreIPCAuditToken>& auditToken)
        : m_arguments(processIdentifier, pageIdentifier, auditToken)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::ProcessIdentifier&, const WebCore::PageIdentifier&, const std::optional<WebKit::CoreIPCAuditToken>&> m_arguments;
};
#endif

#if PLATFORM(VISION) && ENABLE(MODEL_PROCESS)
class RequestSharedSimulationConnection {
public:
    using Arguments = std::tuple<WebKit::CoreIPCAuditToken>;

    static IPC::MessageName name() { return IPC::MessageName::GPUProcess_RequestSharedSimulationConnection; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::GPUProcess_RequestSharedSimulationConnectionReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<std::optional<IPC::SharedFileHandle>>;
    using Reply = CompletionHandler<void(std::optional<IPC::SharedFileHandle>&&)>;
    using Promise = WTF::NativePromise<std::optional<IPC::SharedFileHandle>, IPC::Error>;
    explicit RequestSharedSimulationConnection(const WebKit::CoreIPCAuditToken& modelProcessAuditToken)
        : m_arguments(modelProcessAuditToken)
    {
        ASSERT(!isInAuxiliaryProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::CoreIPCAuditToken&> m_arguments;
};
#endif

} // namespace GPUProcess
} // namespace Messages

#endif // ENABLE(GPU_PROCESS)
