/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "LayerHostingContext.h"
#include "MessageNames.h"
#include "RemoteMediaResourceIdentifier.h"
#include <WebCore/Cookie.h>
#include <WebCore/PlatformMediaResourceLoader.h>
#include <WebCore/ResourceError.h>
#include <WebCore/TrackBase.h>
#include <optional>
#include <span>
#include <wtf/Expected.h>
#include <wtf/Forward.h>
#include <wtf/MediaTime.h>
#include <wtf/RuntimeApplicationChecks.h>
#include <wtf/Seconds.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class FloatSize;
class ISOWebVTTCue;
class ResourceRequest;
class SerializedPlatformDataCueValue;
enum class MediaPlayerReadyState : uint8_t;
struct GenericCueData;
struct VideoFrameMetadata;
struct VideoPlaybackQualityMetrics;
}

namespace WebKit {
struct AudioTrackPrivateRemoteConfiguration;
struct MediaTimeUpdateData;
struct RemoteMediaPlayerState;
struct RemoteVideoFrameProxyProperties;
struct TextTrackPrivateRemoteConfiguration;
struct VideoTrackPrivateRemoteConfiguration;
}

namespace Messages {
namespace MediaPlayerPrivateRemote {

static inline IPC::ReceiverName messageReceiverName()
{
#if ASSERT_ENABLED
    static std::once_flag onceFlag;
    std::call_once(
        onceFlag,
        [&] {
            ASSERT(isInWebProcess());
        }
    );
#endif
    return IPC::ReceiverName::MediaPlayerPrivateRemote;
}

class NetworkStateChanged {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_NetworkStateChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit NetworkStateChanged(const WebKit::RemoteMediaPlayerState& state)
        : m_arguments(state)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&> m_arguments;
};

class ReadyStateChanged {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState, WebCore::MediaPlayerReadyState>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ReadyStateChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ReadyStateChanged(const WebKit::RemoteMediaPlayerState& state, WebCore::MediaPlayerReadyState readyStat)
        : m_arguments(state, readyStat)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&, WebCore::MediaPlayerReadyState> m_arguments;
};

class FirstVideoFrameAvailable {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_FirstVideoFrameAvailable; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    FirstVideoFrameAvailable()
        : m_arguments()
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class VolumeChanged {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_VolumeChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit VolumeChanged(double volume)
        : m_arguments(volume)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<double> m_arguments;
};

class MuteChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_MuteChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit MuteChanged(bool mute)
        : m_arguments(mute)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};

class Seeked {
public:
    using Arguments = std::tuple<WebKit::MediaTimeUpdateData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_Seeked; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit Seeked(const WebKit::MediaTimeUpdateData& timeData)
        : m_arguments(timeData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::MediaTimeUpdateData&> m_arguments;
};

class TimeChanged {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState, WebKit::MediaTimeUpdateData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_TimeChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    TimeChanged(const WebKit::RemoteMediaPlayerState& state, const WebKit::MediaTimeUpdateData& timeData)
        : m_arguments(state, timeData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&, const WebKit::MediaTimeUpdateData&> m_arguments;
};

class DurationChanged {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_DurationChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit DurationChanged(const WebKit::RemoteMediaPlayerState& state)
        : m_arguments(state)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&> m_arguments;
};

class RateChanged {
public:
    using Arguments = std::tuple<double, WebKit::MediaTimeUpdateData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RateChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RateChanged(double rate, const WebKit::MediaTimeUpdateData& timeData)
        : m_arguments(rate, timeData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<double, const WebKit::MediaTimeUpdateData&> m_arguments;
};

class PlaybackStateChanged {
public:
    using Arguments = std::tuple<bool, WebKit::MediaTimeUpdateData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_PlaybackStateChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    PlaybackStateChanged(bool paused, const WebKit::MediaTimeUpdateData& timeData)
        : m_arguments(paused, timeData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool, const WebKit::MediaTimeUpdateData&> m_arguments;
};

class EngineFailedToLoad {
public:
    using Arguments = std::tuple<int64_t>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_EngineFailedToLoad; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit EngineFailedToLoad(int64_t platformErrorCode)
        : m_arguments(platformErrorCode)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<int64_t> m_arguments;
};

class UpdateCachedState {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_UpdateCachedState; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit UpdateCachedState(const WebKit::RemoteMediaPlayerState& state)
        : m_arguments(state)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&> m_arguments;
};

class UpdatePlaybackQualityMetrics {
public:
    using Arguments = std::tuple<WebCore::VideoPlaybackQualityMetrics>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_UpdatePlaybackQualityMetrics; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit UpdatePlaybackQualityMetrics(const WebCore::VideoPlaybackQualityMetrics& metrics)
        : m_arguments(metrics)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::VideoPlaybackQualityMetrics&> m_arguments;
};

class CharacteristicChanged {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaPlayerState>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_CharacteristicChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit CharacteristicChanged(const WebKit::RemoteMediaPlayerState& state)
        : m_arguments(state)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaPlayerState&> m_arguments;
};

class SizeChanged {
public:
    using Arguments = std::tuple<WebCore::FloatSize>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_SizeChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit SizeChanged(const WebCore::FloatSize& naturalSize)
        : m_arguments(naturalSize)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::FloatSize&> m_arguments;
};

class RenderingModeChanged {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RenderingModeChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RenderingModeChanged()
        : m_arguments()
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class CurrentTimeChanged {
public:
    using Arguments = std::tuple<WebKit::MediaTimeUpdateData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_CurrentTimeChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit CurrentTimeChanged(const WebKit::MediaTimeUpdateData& timeData)
        : m_arguments(timeData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::MediaTimeUpdateData&> m_arguments;
};

class AddRemoteAudioTrack {
public:
    using Arguments = std::tuple<WebKit::AudioTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddRemoteAudioTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit AddRemoteAudioTrack(const WebKit::AudioTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::AudioTrackPrivateRemoteConfiguration&> m_arguments;
};

class RemoveRemoteAudioTrack {
public:
    using Arguments = std::tuple<WebCore::TrackID>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveRemoteAudioTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveRemoteAudioTrack(WebCore::TrackID trackID)
        : m_arguments(trackID)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID> m_arguments;
};

class RemoteAudioTrackConfigurationChanged {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebKit::AudioTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoteAudioTrackConfigurationChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RemoteAudioTrackConfigurationChanged(WebCore::TrackID trackID, const WebKit::AudioTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(trackID, configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebKit::AudioTrackPrivateRemoteConfiguration&> m_arguments;
};

class AddRemoteVideoTrack {
public:
    using Arguments = std::tuple<WebKit::VideoTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddRemoteVideoTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit AddRemoteVideoTrack(const WebKit::VideoTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::VideoTrackPrivateRemoteConfiguration&> m_arguments;
};

class RemoveRemoteVideoTrack {
public:
    using Arguments = std::tuple<WebCore::TrackID>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveRemoteVideoTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveRemoteVideoTrack(WebCore::TrackID trackID)
        : m_arguments(trackID)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID> m_arguments;
};

class RemoteVideoTrackConfigurationChanged {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebKit::VideoTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoteVideoTrackConfigurationChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RemoteVideoTrackConfigurationChanged(WebCore::TrackID trackID, const WebKit::VideoTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(trackID, configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebKit::VideoTrackPrivateRemoteConfiguration&> m_arguments;
};

class AddRemoteTextTrack {
public:
    using Arguments = std::tuple<WebKit::TextTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddRemoteTextTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit AddRemoteTextTrack(const WebKit::TextTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::TextTrackPrivateRemoteConfiguration&> m_arguments;
};

class RemoveRemoteTextTrack {
public:
    using Arguments = std::tuple<WebCore::TrackID>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveRemoteTextTrack; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveRemoteTextTrack(WebCore::TrackID trackID)
        : m_arguments(trackID)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID> m_arguments;
};

class RemoteTextTrackConfigurationChanged {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebKit::TextTrackPrivateRemoteConfiguration>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoteTextTrackConfigurationChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RemoteTextTrackConfigurationChanged(WebCore::TrackID trackID, const WebKit::TextTrackPrivateRemoteConfiguration& configuration)
        : m_arguments(trackID, configuration)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebKit::TextTrackPrivateRemoteConfiguration&> m_arguments;
};

class ParseWebVTTFileHeader {
public:
    using Arguments = std::tuple<WebCore::TrackID, String>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ParseWebVTTFileHeader; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ParseWebVTTFileHeader(WebCore::TrackID trackID, const String& header)
        : m_arguments(trackID, header)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const String&> m_arguments;
};

class ParseWebVTTCueData {
public:
    using Arguments = std::tuple<WebCore::TrackID, std::span<const uint8_t>>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueData; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ParseWebVTTCueData(WebCore::TrackID trackID, const std::span<const uint8_t>& data)
        : m_arguments(trackID, data)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const std::span<const uint8_t>&> m_arguments;
};

class ParseWebVTTCueDataStruct {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebCore::ISOWebVTTCue>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueDataStruct; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ParseWebVTTCueDataStruct(WebCore::TrackID trackID, const WebCore::ISOWebVTTCue& cueData)
        : m_arguments(trackID, cueData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebCore::ISOWebVTTCue&> m_arguments;
};

class AddDataCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, MediaTime, MediaTime, std::span<const uint8_t>>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddDataCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    AddDataCue(WebCore::TrackID trackID, const MediaTime& start, const MediaTime& end, const std::span<const uint8_t>& data)
        : m_arguments(trackID, start, end, data)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const MediaTime&, const MediaTime&, const std::span<const uint8_t>&> m_arguments;
};

#if ENABLE(DATACUE_VALUE)
class AddDataCueWithType {
public:
    using Arguments = std::tuple<WebCore::TrackID, MediaTime, MediaTime, WebCore::SerializedPlatformDataCueValue, String>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddDataCueWithType; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    AddDataCueWithType(WebCore::TrackID trackID, const MediaTime& start, const MediaTime& end, const WebCore::SerializedPlatformDataCueValue& cueData, const String& type)
        : m_arguments(trackID, start, end, cueData, type)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const MediaTime&, const MediaTime&, const WebCore::SerializedPlatformDataCueValue&, const String&> m_arguments;
};
#endif

#if ENABLE(DATACUE_VALUE)
class UpdateDataCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, MediaTime, MediaTime, WebCore::SerializedPlatformDataCueValue>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_UpdateDataCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    UpdateDataCue(WebCore::TrackID trackID, const MediaTime& start, const MediaTime& end, const WebCore::SerializedPlatformDataCueValue& cueData)
        : m_arguments(trackID, start, end, cueData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const MediaTime&, const MediaTime&, const WebCore::SerializedPlatformDataCueValue&> m_arguments;
};
#endif

#if ENABLE(DATACUE_VALUE)
class RemoveDataCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, MediaTime, MediaTime, WebCore::SerializedPlatformDataCueValue>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveDataCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RemoveDataCue(WebCore::TrackID trackID, const MediaTime& start, const MediaTime& end, const WebCore::SerializedPlatformDataCueValue& cueData)
        : m_arguments(trackID, start, end, cueData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const MediaTime&, const MediaTime&, const WebCore::SerializedPlatformDataCueValue&> m_arguments;
};
#endif

class AddGenericCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebCore::GenericCueData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_AddGenericCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    AddGenericCue(WebCore::TrackID trackID, const WebCore::GenericCueData& cue)
        : m_arguments(trackID, cue)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebCore::GenericCueData&> m_arguments;
};

class UpdateGenericCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebCore::GenericCueData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_UpdateGenericCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    UpdateGenericCue(WebCore::TrackID trackID, const WebCore::GenericCueData& cue)
        : m_arguments(trackID, cue)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebCore::GenericCueData&> m_arguments;
};

class RemoveGenericCue {
public:
    using Arguments = std::tuple<WebCore::TrackID, WebCore::GenericCueData>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveGenericCue; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RemoveGenericCue(WebCore::TrackID trackID, const WebCore::GenericCueData& cue)
        : m_arguments(trackID, cue)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::TrackID, const WebCore::GenericCueData&> m_arguments;
};

class RequestResource {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaResourceIdentifier, WebCore::ResourceRequest, WebCore::PlatformMediaResourceLoader::LoadOptions>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RequestResource; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    RequestResource(const WebKit::RemoteMediaResourceIdentifier& remoteMediaResourceIdentifier, const WebCore::ResourceRequest& request, const WebCore::PlatformMediaResourceLoader::LoadOptions& options)
        : m_arguments(remoteMediaResourceIdentifier, request, options)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaResourceIdentifier&, const WebCore::ResourceRequest&, const WebCore::PlatformMediaResourceLoader::LoadOptions&> m_arguments;
};

class RemoveResource {
public:
    using Arguments = std::tuple<WebKit::RemoteMediaResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_RemoveResource; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit RemoveResource(const WebKit::RemoteMediaResourceIdentifier& remoteMediaResourceIdentifier)
        : m_arguments(remoteMediaResourceIdentifier)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteMediaResourceIdentifier&> m_arguments;
};

class SendH2Ping {
public:
    using Arguments = std::tuple<URL>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_SendH2Ping; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::MediaPlayerPrivateRemote_SendH2PingReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<Expected<Seconds, WebCore::ResourceError>>;
    using Reply = CompletionHandler<void(Expected<Seconds, WebCore::ResourceError>&&)>;
    using Promise = WTF::NativePromise<Expected<Seconds, WebCore::ResourceError>, IPC::Error>;
    explicit SendH2Ping(const URL& url)
        : m_arguments(url)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const URL&> m_arguments;
};

class ResourceNotSupported {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ResourceNotSupported; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ResourceNotSupported()
        : m_arguments()
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class ActiveSourceBuffersChanged {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_ActiveSourceBuffersChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    ActiveSourceBuffersChanged()
        : m_arguments()
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

#if ENABLE(ENCRYPTED_MEDIA)
class WaitingForKeyChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_WaitingForKeyChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit WaitingForKeyChanged(bool waitingForKey)
        : m_arguments(waitingForKey)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};
#endif

#if ENABLE(ENCRYPTED_MEDIA)
class InitializationDataEncountered {
public:
    using Arguments = std::tuple<String, std::span<const uint8_t>>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_InitializationDataEncountered; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    InitializationDataEncountered(const String& initDataType, const std::span<const uint8_t>& initData)
        : m_arguments(initDataType, initData)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const String&, const std::span<const uint8_t>&> m_arguments;
};
#endif

#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
class MediaPlayerKeyNeeded {
public:
    using Arguments = std::tuple<std::span<const uint8_t>>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_MediaPlayerKeyNeeded; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit MediaPlayerKeyNeeded(const std::span<const uint8_t>& message)
        : m_arguments(message)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const std::span<const uint8_t>&> m_arguments;
};
#endif

#if ENABLE(WIRELESS_PLAYBACK_TARGET)
class CurrentPlaybackTargetIsWirelessChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_CurrentPlaybackTargetIsWirelessChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    explicit CurrentPlaybackTargetIsWirelessChanged(bool isCurrentPlaybackTargetWireless)
        : m_arguments(isCurrentPlaybackTargetWireless)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<bool> m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GetRawCookies {
public:
    using Arguments = std::tuple<URL>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_GetRawCookies; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::MediaPlayerPrivateRemote_GetRawCookiesReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<Vector<WebCore::Cookie>>;
    using Reply = CompletionHandler<void(Vector<WebCore::Cookie>&&)>;
    using Promise = WTF::NativePromise<Vector<WebCore::Cookie>, IPC::Error>;
    explicit GetRawCookies(const URL& url)
        : m_arguments(url)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const URL&> m_arguments;
};
#endif

#if PLATFORM(COCOA)
class PushVideoFrameMetadata {
public:
    using Arguments = std::tuple<WebCore::VideoFrameMetadata, WebKit::RemoteVideoFrameProxyProperties>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_PushVideoFrameMetadata; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    PushVideoFrameMetadata(const WebCore::VideoFrameMetadata& metadata, const WebKit::RemoteVideoFrameProxyProperties& frameProperties)
        : m_arguments(metadata, frameProperties)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::VideoFrameMetadata&, const WebKit::RemoteVideoFrameProxyProperties&> m_arguments;
};
#endif

#if PLATFORM(COCOA)
class LayerHostingContextIdChanged {
public:
    using Arguments = std::tuple<std::optional<WebKit::LayerHostingContextID>, WebCore::FloatSize>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_LayerHostingContextIdChanged; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    LayerHostingContextIdChanged(const std::optional<WebKit::LayerHostingContextID>& inlineLayerHostingContextId, const WebCore::FloatSize& presentationSize)
        : m_arguments(inlineLayerHostingContextId, presentationSize)
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const std::optional<WebKit::LayerHostingContextID>&, const WebCore::FloatSize&> m_arguments;
};
#endif

class CommitAllTransactions {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::MediaPlayerPrivateRemote_CommitAllTransactions; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::MediaPlayerPrivateRemote_CommitAllTransactionsReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<>;
    using Reply = CompletionHandler<void()>;
    using Promise = WTF::NativePromise<void, IPC::Error>;
    CommitAllTransactions()
        : m_arguments()
    {
        ASSERT(isInGPUProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

} // namespace MediaPlayerPrivateRemote
} // namespace Messages

#endif // ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
