/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(GPU_PROCESS)

#include "ArgumentCoders.h"
#include "BufferIdentifierSet.h"
#include "Connection.h"
#include "IPCSemaphore.h"
#include "MarkSurfacesAsVolatileRequestIdentifier.h"
#include "MessageNames.h"
#include "PrepareBackingStoreBuffersData.h"
#include "RemoteImageBufferSetIdentifier.h"
#include "RenderingUpdateID.h"
#include "ShapeDetectionIdentifier.h"
#include "SwapBuffersDisplayRequirement.h"
#include <WebCore/BarcodeDetectorOptionsInterface.h>
#include <WebCore/BarcodeFormatInterface.h>
#include <WebCore/DecomposedGlyphs.h>
#include <WebCore/DestinationColorSpace.h>
#include <WebCore/FaceDetectorOptionsInterface.h>
#include <WebCore/Filter.h>
#include <WebCore/Gradient.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/RenderingMode.h>
#include <WebCore/RenderingResourceIdentifier.h>
#include <WebCore/ShareableBitmapHandle.h>
#include <WebCore/SnapshotIdentifier.h>
#include <optional>
#include <utility>
#include <wtf/Forward.h>
#include <wtf/OptionSet.h>
#include <wtf/Ref.h>
#include <wtf/RuntimeApplicationChecks.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>

namespace WebCore {
class FloatSize;
enum class ImageBufferPixelFormat : uint8_t;
struct FontCustomPlatformSerializedData;
struct FontInternalAttributes;
struct FontPlatformDataAttributes;
struct ImageBufferResourceLimits;
}

namespace Messages {
namespace RemoteRenderingBackend {

static inline IPC::ReceiverName messageReceiverName()
{
#if ASSERT_ENABLED
    static std::once_flag onceFlag;
    std::call_once(
        onceFlag,
        [&] {
            ASSERT(isInGPUProcess());
        }
    );
#endif
    return IPC::ReceiverName::RemoteRenderingBackend;
}

class CreateImageBuffer {
public:
    using Arguments = std::tuple<WebCore::FloatSize, WebCore::RenderingMode, WebCore::RenderingPurpose, float, WebCore::DestinationColorSpace, WebCore::ImageBufferPixelFormat, WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CreateImageBuffer; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    CreateImageBuffer(const WebCore::FloatSize& logicalSize, const WebCore::RenderingMode& renderingMode, const WebCore::RenderingPurpose& renderingPurpose, float resolutionScale, const WebCore::DestinationColorSpace& colorSpace, WebCore::ImageBufferPixelFormat pixelFormat, const WebCore::RenderingResourceIdentifier& renderingResourceIdentifier)
        : m_arguments(logicalSize, renderingMode, renderingPurpose, resolutionScale, colorSpace, pixelFormat, renderingResourceIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::FloatSize&, const WebCore::RenderingMode&, const WebCore::RenderingPurpose&, float, const WebCore::DestinationColorSpace&, WebCore::ImageBufferPixelFormat, const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class ReleaseImageBuffer {
public:
    using Arguments = std::tuple<WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseImageBuffer; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseImageBuffer(const WebCore::RenderingResourceIdentifier& renderingResourceIdentifier)
        : m_arguments(renderingResourceIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class GetImageBufferResourceLimitsForTesting {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_GetImageBufferResourceLimitsForTesting; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isReplyStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::RemoteRenderingBackend_GetImageBufferResourceLimitsForTestingReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<WebCore::ImageBufferResourceLimits>;
    using Reply = CompletionHandler<void(WebCore::ImageBufferResourceLimits&&)>;
    using Promise = WTF::NativePromise<WebCore::ImageBufferResourceLimits, IPC::Error>;
    GetImageBufferResourceLimitsForTesting()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class DestroyGetPixelBufferSharedMemory {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_DestroyGetPixelBufferSharedMemory; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    DestroyGetPixelBufferSharedMemory()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class CacheNativeImage {
public:
    using Arguments = std::tuple<WebCore::ShareableBitmapHandle, WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheNativeImage; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    CacheNativeImage(WebCore::ShareableBitmapHandle&& handle, const WebCore::RenderingResourceIdentifier& renderingResourceIdentifier)
        : m_arguments(WTFMove(handle), renderingResourceIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<WebCore::ShareableBitmapHandle&&, const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class CacheFont {
public:
    using Arguments = std::tuple<WebCore::FontInternalAttributes, WebCore::FontPlatformDataAttributes, std::optional<WebCore::RenderingResourceIdentifier>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheFont; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    CacheFont(const WebCore::FontInternalAttributes& data, const WebCore::FontPlatformDataAttributes& platformData, const std::optional<WebCore::RenderingResourceIdentifier>& renderingResourceIdentifier)
        : m_arguments(data, platformData, renderingResourceIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::FontInternalAttributes&, const WebCore::FontPlatformDataAttributes&, const std::optional<WebCore::RenderingResourceIdentifier>&> m_arguments;
};

class CacheFontCustomPlatformData {
public:
    using Arguments = std::tuple<WebCore::FontCustomPlatformSerializedData>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheFontCustomPlatformData; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    explicit CacheFontCustomPlatformData(const WebCore::FontCustomPlatformSerializedData& fontCustomPlatformData)
        : m_arguments(fontCustomPlatformData)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::FontCustomPlatformSerializedData&> m_arguments;
};

class CacheDecomposedGlyphs {
public:
    using Arguments = std::tuple<Ref<WebCore::DecomposedGlyphs>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheDecomposedGlyphs; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit CacheDecomposedGlyphs(const Ref<WebCore::DecomposedGlyphs>& decomposedGlyphs)
        : m_arguments(decomposedGlyphs)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Ref<WebCore::DecomposedGlyphs>&> m_arguments;
};

class CacheGradient {
public:
    using Arguments = std::tuple<Ref<WebCore::Gradient>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheGradient; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    explicit CacheGradient(const Ref<WebCore::Gradient>& gradient)
        : m_arguments(gradient)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Ref<WebCore::Gradient>&> m_arguments;
};

class CacheFilter {
public:
    using Arguments = std::tuple<Ref<WebCore::Filter>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CacheFilter; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    explicit CacheFilter(const Ref<WebCore::Filter>& filter)
        : m_arguments(filter)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Ref<WebCore::Filter>&> m_arguments;
};

class ReleaseAllDrawingResources {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseAllDrawingResources; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    ReleaseAllDrawingResources()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class ReleaseAllImageResources {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseAllImageResources; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    ReleaseAllImageResources()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class ReleaseRenderingResource {
public:
    using Arguments = std::tuple<WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseRenderingResource; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseRenderingResource(const WebCore::RenderingResourceIdentifier& renderingResourceIdentifier)
        : m_arguments(renderingResourceIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class CreateRemoteImageBufferSet {
public:
    using Arguments = std::tuple<WebKit::RemoteImageBufferSetIdentifier, WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CreateRemoteImageBufferSet; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    CreateRemoteImageBufferSet(const WebKit::RemoteImageBufferSetIdentifier& bufferSetIdentifier, const WebCore::RenderingResourceIdentifier& displayListIdentifier)
        : m_arguments(bufferSetIdentifier, displayListIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteImageBufferSetIdentifier&, const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class ReleaseRemoteImageBufferSet {
public:
    using Arguments = std::tuple<WebKit::RemoteImageBufferSetIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseRemoteImageBufferSet; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseRemoteImageBufferSet(const WebKit::RemoteImageBufferSetIdentifier& bufferSetIdentifier)
        : m_arguments(bufferSetIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RemoteImageBufferSetIdentifier&> m_arguments;
};

#if PLATFORM(COCOA)
class PrepareImageBufferSetsForDisplay {
public:
    using Arguments = std::tuple<Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_PrepareImageBufferSetsForDisplay; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit PrepareImageBufferSetsForDisplay(const Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>& swapBuffersInput)
        : m_arguments(swapBuffersInput)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>&> m_arguments;
};
#endif

#if PLATFORM(COCOA)
class PrepareImageBufferSetsForDisplaySync {
public:
    using Arguments = std::tuple<Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_PrepareImageBufferSetsForDisplaySync; }
    static constexpr bool isSync = true;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isReplyStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<Vector<WebKit::SwapBuffersDisplayRequirement>>;
    using Reply = CompletionHandler<void(Vector<WebKit::SwapBuffersDisplayRequirement>&&)>;
    explicit PrepareImageBufferSetsForDisplaySync(const Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>& swapBuffersInput)
        : m_arguments(swapBuffersInput)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const Vector<WebKit::ImageBufferSetPrepareBufferForDisplayInputData>&> m_arguments;
};
#endif

class MarkSurfacesVolatile {
public:
    using Arguments = std::tuple<WebKit::MarkSurfacesAsVolatileRequestIdentifier, Vector<std::pair<WebKit::RemoteImageBufferSetIdentifier, OptionSet<WebKit::BufferInSetType>>>, bool>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_MarkSurfacesVolatile; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    MarkSurfacesVolatile(const WebKit::MarkSurfacesAsVolatileRequestIdentifier& requestIdentifier, const Vector<std::pair<WebKit::RemoteImageBufferSetIdentifier, OptionSet<WebKit::BufferInSetType>>>& renderingResourceIdentifiers, bool forcePurge)
        : m_arguments(requestIdentifier, renderingResourceIdentifiers, forcePurge)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::MarkSurfacesAsVolatileRequestIdentifier&, const Vector<std::pair<WebKit::RemoteImageBufferSetIdentifier, OptionSet<WebKit::BufferInSetType>>>&, bool> m_arguments;
};

class FinalizeRenderingUpdate {
public:
    using Arguments = std::tuple<WebKit::RenderingUpdateID>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_FinalizeRenderingUpdate; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit FinalizeRenderingUpdate(const WebKit::RenderingUpdateID& renderingUpdateID)
        : m_arguments(renderingUpdateID)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::RenderingUpdateID&> m_arguments;
};

#if USE(GRAPHICS_LAYER_WC)
class Flush {
public:
    using Arguments = std::tuple<IPC::Semaphore>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_Flush; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = false;
    static constexpr bool isStreamBatched = false;

    explicit Flush(const IPC::Semaphore& semaphore)
        : m_arguments(semaphore)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const IPC::Semaphore&> m_arguments;
};
#endif

class MoveToSerializedBuffer {
public:
    using Arguments = std::tuple<WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_MoveToSerializedBuffer; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit MoveToSerializedBuffer(const WebCore::RenderingResourceIdentifier& sourceImageBuffer)
        : m_arguments(sourceImageBuffer)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::RenderingResourceIdentifier&> m_arguments;
};

class MoveToImageBuffer {
public:
    using Arguments = std::tuple<WebCore::RenderingResourceIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_MoveToImageBuffer; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit MoveToImageBuffer(const WebCore::RenderingResourceIdentifier& destinationImageBuffer)
        : m_arguments(destinationImageBuffer)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::RenderingResourceIdentifier&> m_arguments;
};

#if PLATFORM(COCOA)
class DidDrawRemoteToPDF {
public:
    using Arguments = std::tuple<WebCore::PageIdentifier, WebCore::RenderingResourceIdentifier, WebCore::SnapshotIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_DidDrawRemoteToPDF; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    DidDrawRemoteToPDF(const WebCore::PageIdentifier& pageID, const WebCore::RenderingResourceIdentifier& imageBufferIdentifier, const WebCore::SnapshotIdentifier& snapshotIdentifier)
        : m_arguments(pageID, imageBufferIdentifier, snapshotIdentifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebCore::PageIdentifier&, const WebCore::RenderingResourceIdentifier&, const WebCore::SnapshotIdentifier&> m_arguments;
};
#endif

class CreateRemoteBarcodeDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier, WebCore::ShapeDetection::BarcodeDetectorOptions>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CreateRemoteBarcodeDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    CreateRemoteBarcodeDetector(const WebKit::ShapeDetectionIdentifier& identifier, const WebCore::ShapeDetection::BarcodeDetectorOptions& barcodeDetectorOptions)
        : m_arguments(identifier, barcodeDetectorOptions)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&, const WebCore::ShapeDetection::BarcodeDetectorOptions&> m_arguments;
};

class ReleaseRemoteBarcodeDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseRemoteBarcodeDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseRemoteBarcodeDetector(const WebKit::ShapeDetectionIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&> m_arguments;
};

class GetRemoteBarcodeDetectorSupportedFormats {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_GetRemoteBarcodeDetectorSupportedFormats; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isReplyStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::RemoteRenderingBackend_GetRemoteBarcodeDetectorSupportedFormatsReply; }
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using ReplyArguments = std::tuple<Vector<WebCore::ShapeDetection::BarcodeFormat>>;
    using Reply = CompletionHandler<void(Vector<WebCore::ShapeDetection::BarcodeFormat>&&)>;
    using Promise = WTF::NativePromise<Vector<WebCore::ShapeDetection::BarcodeFormat>, IPC::Error>;
    GetRemoteBarcodeDetectorSupportedFormats()
        : m_arguments()
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<> m_arguments;
};

class CreateRemoteFaceDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier, WebCore::ShapeDetection::FaceDetectorOptions>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CreateRemoteFaceDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    CreateRemoteFaceDetector(const WebKit::ShapeDetectionIdentifier& identifier, const WebCore::ShapeDetection::FaceDetectorOptions& faceDetectorOptions)
        : m_arguments(identifier, faceDetectorOptions)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&, const WebCore::ShapeDetection::FaceDetectorOptions&> m_arguments;
};

class ReleaseRemoteFaceDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseRemoteFaceDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseRemoteFaceDetector(const WebKit::ShapeDetectionIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&> m_arguments;
};

class CreateRemoteTextDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_CreateRemoteTextDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit CreateRemoteTextDetector(const WebKit::ShapeDetectionIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&> m_arguments;
};

class ReleaseRemoteTextDetector {
public:
    using Arguments = std::tuple<WebKit::ShapeDetectionIdentifier>;

    static IPC::MessageName name() { return IPC::MessageName::RemoteRenderingBackend_ReleaseRemoteTextDetector; }
    static constexpr bool isSync = false;
    static constexpr bool canDispatchOutOfOrder = false;
    static constexpr bool replyCanDispatchOutOfOrder = false;
    static constexpr bool deferSendingIfSuspended = false;
    static constexpr bool isStreamEncodable = true;
    static constexpr bool isStreamBatched = false;

    explicit ReleaseRemoteTextDetector(const WebKit::ShapeDetectionIdentifier& identifier)
        : m_arguments(identifier)
    {
        ASSERT(isInWebProcess());
    }

    auto&& arguments()
    {
        return WTFMove(m_arguments);
    }

private:
    std::tuple<const WebKit::ShapeDetectionIdentifier&> m_arguments;
};

} // namespace RemoteRenderingBackend
} // namespace Messages

#endif // ENABLE(GPU_PROCESS)
